@@ -1,5 +1,5 @@
 from .core import *
 from .event import event
-from .__array import array
+from .rparray import array
 from .plotting import *
 from .functionality import *
\ No newline at end of file
