@@ -3,7 +3,7 @@
 #############################
 
 from .core import *
-from .__array import array
+from .rparray import array
 from .event import *
 
 # Print the closest lightnings
@@ -251,15 +251,20 @@ def get_dict(filename:str,struct=ORBIT_STRUCT,condition:str=None,MAX=None,STUPID
         except:
             print(bcolors.WARNING+'WARNING!' + bcolors.ENDC +' Condition ' + condition + ' is not valid for the dataset you requested. The data returned will not be filtered')
 
+    # Specific loading changes
+    if 'temperature' in struct.keys():
+        data['temperature'] -= 55
+        
     # Return the dictionary
     return data
 
 # Corrects the timestamp based on orbit rate
-def correct_time_orbit(orbit:dict,TIME:int=20,RANGE=(0,100)):
+def correct_time_orbit(orbit:dict,key:str='rate0',TIME:int=20,RANGE=(0,100)):
     """Corrects the time of events based on the data of the orbit buffer
 
     Args:
         orbit (dict): The orbit buffer
+        key (str): Key for the corresponding event buffer rate
         TIME (int, optional): The period of the rate measurements. Defaults to 20.
         RANGE (tuple, optional): a range of indices to translate of the orbit buffer. Defaults to (0,100).
 
@@ -275,14 +280,14 @@ def correct_time_orbit(orbit:dict,TIME:int=20,RANGE=(0,100)):
 
     # Start counting events from the correct timestamp
     if RANGE[0] != 0:
-        for counts in orbit['ratev'][0:RANGE[0]]:
+        for counts in orbit[key][0:RANGE[0]]:
             start_cnt += int(counts * TIME)
         
         # Start counting from this value
         end_cnt += start_cnt
 
     # For each count in the orbit
-    for count in orbit['ratev'][RANGE[0]:RANGE[1]]:
+    for count in orbit[key][RANGE[0]:RANGE[1]]:
         # Get the next number of counts
         count = int(count*TIME)
         if count == 0:
@@ -376,12 +381,13 @@ def correct_time_FPGA(data:dict,RIZE_TIME:float=1,CONST_TIME:float=1,TMAX:int=10
     return timestamp, valid_data
 
 # Now putting everything together
-def correct_time(data:dict,orbit:dict,TIME:int=20,RANGE_ORBIT=(0,100),RIZE_TIME:float=1,CONST_TIME:float=1,TMAX:int=10000-1):
+def correct_time(data:dict,orbit:dict,key:str='rate0',TIME:int=20,RANGE_ORBIT=(0,100),RIZE_TIME:float=1,CONST_TIME:float=1,TMAX:int=10000-1):
     """Correct time using both FPGA and Orbit corrections simultaneously and generate a timestamp for the valid_data
 
     Args:
         data (dict): The data buffer to correct the timestamp of
         orbit (dict): The corresponding orbit buffer
+        key (str): Key for the corresponding event buffer rate
         TIME (int, optional): Period of the orbit buffer measurments. Defaults to 20.
         RANGE_ORBIT (tuple, optional): The range of indices in the orbit buffer to translate. Defaults to (0,100).
         RIZE_TIME (int, optional): The time it takes for the FPGA counter to saturate. Defaults to 1.
@@ -403,13 +409,13 @@ def correct_time(data:dict,orbit:dict,TIME:int=20,RANGE_ORBIT=(0,100),RIZE_TIME:
 
     # Start counting events from the correct timestamp
     if RANGE_ORBIT[0] != 0:
-        for counts in orbit['ratev'][0:RANGE_ORBIT[0]]:
+        for counts in orbit[key][0:RANGE_ORBIT[0]]:
             processed_cnt += int(counts * TIME)
 
     # Error flag
     oops = 0
     # For each count in the orbit
-    for count in orbit['ratev'][RANGE_ORBIT[0]:RANGE_ORBIT[1]]:
+    for count in orbit[key][RANGE_ORBIT[0]:RANGE_ORBIT[1]]:
         # Get the next number of counts
         count = int(count*TIME)
         if count == 0:
@@ -437,7 +443,7 @@ def correct_time(data:dict,orbit:dict,TIME:int=20,RANGE_ORBIT=(0,100),RIZE_TIME:
         total_cnt       += len(valid_data)
         processed_cnt   += count
 
-    print("Oops': ",oops/(RANGE_ORBIT[1]-RANGE_ORBIT[0]))
+    if oops != 0: print("Oops': ",oops/(RANGE_ORBIT[1]-RANGE_ORBIT[0]))
 
     # # remove the last element of the timestamp
     # timestamp = timestamp[:-1]
@@ -446,3 +452,315 @@ def correct_time(data:dict,orbit:dict,TIME:int=20,RANGE_ORBIT=(0,100),RIZE_TIME:
     total_cnt = int(total_cnt)
 
     return timestamp, total_cnt, valid_events
+
+
+# Order the data according to entry number
+def sort(data,field='entry_nr'):
+    """Sort the data based on a metadata field
+
+    Args:
+        data (array of dictionaries): The array of dictionaries from the downloaded data
+        field (str, optional): The metadata field to sort according to. Defaults to 'entry_nr'.
+
+    Returns:
+        sorted: Sorted list of lists
+    """
+    if len(data) <= 1: return data
+    
+    # Get the indices
+    idx = np.argsort([d[field] for d in data])
+    
+    # Sorted array
+    sorted = [data[idx[i]] for i in range(len(data))]
+
+    return sorted
+
+# Download data based on various keys
+def download_file_ver(buffer:int = 1, file_ver=1):
+    """Download a data from NA server with a common file version
+
+    Args:
+        buffer (int, optional): The buffer to download. Defaults to 1.
+        file_ver (int, optional): The file version number. Defaults to 1.
+
+    Returns:
+        data: list of dictionaries with the rows
+    """
+    # Generate some variables
+    fileName="pc_buff"+str(buffer)
+    host=HOST
+    token=TOKEN
+
+    # Create a rest request
+    rest = RestOperations(f'{host}/{fileName}_download?file_ver=eq.{file_ver}', authType = 'token', token = token)
+   
+    # Download the data using the request
+    data = rest.SendGetReq()
+
+    # Sort the data
+    data = sort(data)
+
+    return data
+
+# Download data based on time range
+def download_time_delta(buffer:int = 1, start:str='2022-06-01T00:00:00', end:str='2022-06-07T00:00:00'):
+    """Download NA data on a time interval 
+
+    Args:
+        buffer (int, optional): The buffer number. Defaults to 1.
+        start (str, optional): String with iso date to start. Defaults to '2022-06-01T00:00:00'.
+        end (str, optional): String with iso date to end. Defaults to '2022-06-07T00:00:00'.
+
+    Returns:
+        data: list of dictionaries with the rows
+    """
+    # Generate some variables
+    fileName="pc_buff"+str(buffer)
+    host=HOST
+    token=TOKEN
+    
+    # Create a rest request
+    rest = RestOperations(f'{host}/{fileName}_download?archived_ts=gte.{start}&archived_ts=lt.{end}', authType = 'token', token = token)
+
+    # Download the data using the request
+    data = rest.SendGetReq()
+
+    # Sort the data
+    data = sort(data)
+
+    return data
+
+# Save this data to a file to avoid having them in memory
+def save_raw_data(data,filepath:str='./',buffer:int=1):
+    """Save the raw data to a file in the computer
+
+    Args:
+        data (_type_): The raw data downloaded from NA server
+        filepath (str, optional): The path that you want to save the file to. Defaults to './'.
+        buffer (int, optional): The buffer number. Defaults to 1.
+
+    Returns:
+        string: The filename of the file.
+    """
+    # Create the filename
+    timestamp   = data[0]['archived_ts']
+    date        = timestamp[0:timestamp.index('T')]
+    filename    = filepath + f'Light1_{date}_Buff{buffer}.dat'
+
+    # Load the file to write the output
+    file = open(filename,'wb')
+
+    # Append the data
+    for row in data:
+        # Convert the hexadecimal entry to bytes
+        entry = bytes.fromhex(row['entry_data'][2:])
+        file.write(entry)
+    
+    # Close the file
+    file.close()
+
+    # Return the filename if you need it
+    return filename
+
+# Convert from binary
+def log_to_ascii(data,fileName:str=None):
+    """Decode binary log file to ascii
+
+    Args:
+        data (dictionary): The dictionary obtained from the downloaded NA code
+        fileName (str, optional): Filename to export the logfile to. If None then the file is not exported. Defaults to None.
+
+    Returns:
+        str: The decoded logfile as a string
+    """
+    # Store the full decoded text here
+    full_text = ''
+
+    # For every line in the logfile
+    for entry in data:
+        line =  bytes.fromhex(entry['entry_data'][2:]).decode("ASCII")
+        full_text += line
+
+    # If you need to store do so
+    if fileName is not None: 
+        file = open(fileName,'w')
+        file.write(full_text)
+        file.close()
+
+    # Return the full text
+    return full_text
+
+
+# Parse a command
+def desc_finder(line:str,cmdlist,outputs,i,time,failed_idx):
+    """Parse a command and return its status and description
+
+    Args:
+        line (str): The string of the command
+        cmdlist (pd.DataFrame): pandas data frame with the commands and their equivalent messages
+        outputs (_type_): _description_
+        i (_type_): _description_
+        time (_type_): _description_
+        failed_idx (_type_): _description_
+
+    Returns:
+        _type_: _description_
+    """
+
+    # Get the description
+    status = 1
+    splt = line.split(' ')
+
+    # define end of log file
+    if splt[-1] == 'SE0>':
+        # desc = 'LOG END'
+        desc = [-1,17]
+
+    # define commands from the command file
+    elif 'txrx' in splt[1]:
+        node,port,msg = int(splt[2]),int(splt[3]),str(splt[5])
+
+        index = cmdlist.loc[(cmdlist['NODE']==node) & (cmdlist['PORT']==port) & (cmdlist['Message'].str.startswith(msg)),['ID_COMMAND_Proposed','ID_in_Graph']]
+        
+        # include the power shutdown
+        if node == 4:
+            index = cmdlist.loc[(cmdlist['NODE']==node) & (cmdlist['PORT']==port),['ID_COMMAND_Proposed','ID_in_Graph']]
+
+        # include the custom scenario
+        if port == 9:
+            index = cmdlist.loc[(cmdlist['NODE']==node) & (cmdlist['PORT']==port),['ID_COMMAND_Proposed','ID_in_Graph']]
+        
+
+        #If index did not find anything
+        if len(index) == 0:
+            desc = []
+            
+        # if command found in command list
+        else:
+            desc = list(index.to_numpy()[0])
+        
+
+        if i in [fid for fid in failed_idx]:
+            time = time + (float(splt[4])/1000)
+            status = -1
+        
+
+    elif splt[1] == 'delay':
+        desc = []
+
+    elif splt[1] == 'delayuntil':
+        time = float(splt[2])
+        desc = []
+
+    elif 'read' in splt[1]:
+        time = float(outputs[i][0].split(' ')[3])
+        desc = []
+
+    else:
+        desc = []
+    
+    return desc,time,status
+
+
+
+# Decode a logfile
+def decode_log(filename:str="../../../Data/Logs/light1-SD-1016-se-log.txt"):
+    """Take a log file and parse its commands to uncover thier timestamps
+
+    Args:
+        filename (str, optional): The filepath and filename of the log file. Defaults to "../../../Data/Logs/light1-SD-1016-se-log.txt".
+
+    Returns:
+        _type_: Log, commands, outputs, description, failed_idx, loglines_array
+    """
+
+    # Load the logfile
+    logfile = open(filename)
+    cmdlist = pd.read_csv("command_list.csv")
+
+    # Load the lines
+    loglines = logfile.readlines()
+
+    # Close the file
+    logfile.close()
+
+    # Create an array with the lines
+    commands    = []
+    outputs     = []
+    idx         = []
+    description = []
+    times       = []
+    fails       = []
+    failed_idx  = []
+
+    # Get commands and their indices
+    for i, line in enumerate(loglines):
+        if 'SE0>' in line:
+            commands.append(line)
+            idx.append(i)
+
+    # Check if the last command was empty
+    if loglines[-1] != 'SE0>':
+        commands.append('SE0>')
+        idx.append(len(loglines))
+
+    # Get the command output
+    for i in range(len(idx)-1):
+        out = []
+        if 'SE0>#' not in commands[i]:
+            for j in range(idx[i]+1,idx[i+1]):
+                out.append(loglines[j])
+        
+        outputs.append(out)
+
+    # Create the dictionary
+    log = [[command,output] for command,output in zip(commands,outputs)]
+
+    # Check which commands executed correcly
+  
+
+    # Find the commands that failed
+    for i in range(len(commands)-1):
+        for output in log[i][1]:
+            if 'FAIL' in output:
+                failed_idx.append(i)
+
+
+    # Initialize time at 0s
+    time = 0
+    
+    # Give commands and use the finder function to pull the description, time ran, and if the command failed.
+    for k, id in enumerate(idx):
+        c = commands[k]
+        cmd = c.split('\n')[0]
+        desc,time,failed = desc_finder(cmd,cmdlist,outputs,k,time,failed_idx)
+
+        if len(desc) != 0:
+            description.append(desc)
+            times.append(time)
+            fails.append(failed)
+    tempname = 'templog.csv'
+    fileDir = os.path.join('Defined_Logs')
+
+    if not os.path.exists(fileDir):
+        os.makedirs(fileDir)
+    
+    filePath = os.path.join(fileDir, tempname)
+
+    #Name the headers of the CSV File
+    header = ['status','time','description','ID_in_Graph']
+
+    loglines_array = []
+
+    #Clear the file and write from scratch
+    with open(filePath, 'w', newline='') as file:
+        writer = csv.writer(file)
+        writer.writerow(header)
+
+        for i in range(len(description)):
+            g = (fails[i], times[i-1], description[i][0],description[i][1])
+            loglines_array.append(g)
+            writer.writerow(g)
+
+    # Return everything else
+    return log, commands[:-1], outputs, description, failed_idx, loglines_array
